/*
 *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
 */

/*  This file is part of Kelbt.
 *
 *  Kelbt is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Kelbt is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Kelbt; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

%{

#define YY_NEVER_INTERACTIVE 1
//#define WANT_TOKEN_WRITE

#include <iostream>
#include "kelbt.h"
#include "parsedata.h"
#include "buffer.h"

using std::cout;
using std::cerr;
using std::endl;

Buffer tokbuf;
int builtinBrace = 0;
bool handlingInclude = false;

/* These keep track of the start of an inline comment or literal string for
 * reporting unterminated comments or strings. */
int il_comm_lit_first_line;
int il_comm_lit_first_column;

/* These keep track of the start of a code block for reporting unterminated
 * code blocks. */
int il_code_first_line;
int il_code_first_column;

/* Include Stack data. */
YY_BUFFER_STATE buff_stack[INCLUDE_STACK_SIZE];
int inc_stack_ptr = 0;

YYSTYPE *yylval;
YYLTYPE *yylloc;

extern InputData *id;
extern int includeDepth;

void extendGarble();
void garble();

void extendToken( char *data, int len );
int emitToken( int token, char *data, int len );
int emitNoData( int token );
void passThrough( char *data );
void popInclude();

/* Using a wrapper for the parser, must the lex declaration. */
#define YY_DECL int kelbt_lex()

%}

/* Outside an fsm machine specification ("outside code"). */
%x		OC_SGL_LIT
%x		OC_DBL_LIT
%x		OC_C_COM
%x		OC_CXX_COM

/* Inside a fsm machine specification. */
%x		KL_INITIAL
%x		KL_SLIT
%x		KL_DLIT
%x		KL_SHELL_COM

/* Inline code. */
%x		IL_INITIAL
%x		IL_SGL_LIT
%x		IL_DBL_LIT
%x		IL_C_COM
%x		IL_CXX_COM

WSCHAR [\t\n\v\f\r ]
IDENT [a-zA-Z_][a-zA-Z_0-9]*

%%

	/* Numbers in outter code. */
<INITIAL>[0-9]+ {
	garble();
	passThrough( yytext );
}

	/* Words in outter code. */
<INITIAL>{IDENT} {
	garble();
	passThrough( yytext );
}

	/* Begin a C style comment. */
<INITIAL>"/*" {
	BEGIN(OC_C_COM);
	extendGarble();
	passThrough( yytext );
}
	/* Data in a C style comment. */
<OC_C_COM>. 		extendGarble(); passThrough( yytext );
<OC_C_COM>\n 		extendGarble(); passThrough( yytext );

	/* Terminate a C style comment. */
<OC_C_COM>"*/" {
	BEGIN(INITIAL);
	garble();
	passThrough( yytext );
}

	/* Begin a C++ style comment. */
<INITIAL>"//" {
	BEGIN(OC_CXX_COM);
	extendGarble();
	passThrough( yytext );
}
	/* Data in a C++ style comment. */
<OC_CXX_COM>[^\n]+ {
	extendGarble();
	passThrough( yytext );
}
	/* Terminate a C++ style comment. */
<OC_CXX_COM>\n {
	BEGIN(INITIAL);
	garble();
	passThrough( yytext );
}


	/* Start literals. */
<INITIAL>\' {
	BEGIN(OC_SGL_LIT);
	extendGarble();
	passThrough( yytext );
}
<INITIAL>\" {
	BEGIN(OC_DBL_LIT);
	extendGarble();
	passThrough( yytext );
}
	/* Various escape sequences in literals. We don't need to get them
	 * all here. We just need to pick off the ones that could confuse us
	 * about the literal we are matchine */
<OC_SGL_LIT,OC_DBL_LIT>\\\'		extendGarble(); passThrough( yytext );
<OC_SGL_LIT,OC_DBL_LIT>\\\"		extendGarble(); passThrough( yytext );
<OC_SGL_LIT,OC_DBL_LIT>\\\\		extendGarble(); passThrough( yytext );
	/* Characters in literals. */
<OC_DBL_LIT>[^\"]				extendGarble(); passThrough( yytext );
<OC_SGL_LIT>[^\']				extendGarble(); passThrough( yytext );
	/* Terminate a double literal */
<OC_DBL_LIT>\" {
	BEGIN(INITIAL);
	garble();
	passThrough( yytext );
}
	/* Terminate a single literal. */
<OC_SGL_LIT>\' {
	BEGIN(INITIAL);
	garble();
	passThrough( yytext );
}

	/* Whitespace. */
<INITIAL>{WSCHAR}+ {
	garble();
	passThrough( yytext );
}

	/* Section Deliminator */
<INITIAL>"%%{" {
	BEGIN(KL_INITIAL);
	return emitNoData( TK_StartSection );
}
<KL_INITIAL>"}%%" {
	BEGIN(INITIAL);
	return emitNoData( TK_EndSection );
}

	/* Any other characters. */
<INITIAL>. {
	garble();
	passThrough( yytext );
}

	/* Keywords in RL and IL. */
<KL_INITIAL>init {
	return emitNoData( KW_Init );
}
<KL_INITIAL>interface {
	return emitNoData( KW_Interface );
}
<KL_INITIAL>context {
	return emitNoData( KW_Context );
}
<KL_INITIAL>include {
	/* Include tokens statments are processed by both the scanner and the
	 * parser.  The scanner opens the include file and switches to it and the
	 * parser invokes a new parser for handling the tokens. We use
	 * handlingInclude to indicate that the scanner is processing an include
	 * directive. Ends at ; */
	handlingInclude = true;
	return emitNoData( KW_Include );
}

<KL_INITIAL>struct {
	return emitNoData( KW_Struct );
}
<KL_INITIAL>element {
	return emitNoData( KW_Element );
}
<KL_INITIAL>token {
	return emitNoData( KW_Token );
}
<KL_INITIAL>nonterm {
	return emitNoData( KW_NonTerm );
}
<KL_INITIAL>pre_reduce {
	return emitNoData( KW_PreReduce );
}
<KL_INITIAL>post_reduce {
	return emitNoData( KW_PostReduce );
}
<KL_INITIAL>translate {
	return emitNoData( KW_Translate );
}
<KL_INITIAL>try {
	return emitNoData( KW_Try );
}
<KL_INITIAL>final {
	return emitNoData( KW_Final );
}
<KL_INITIAL>parser {
	return emitNoData( KW_Parser );
}

<KL_INITIAL>type {
	return emitNoData( KW_Type );
}
<KL_INITIAL>uses {
	return emitNoData( KW_Uses );
}
<KL_INITIAL>undo {
	return emitNoData( KW_Undo );
}
<KL_INITIAL>commit {
	return emitNoData( KW_Commit );
}

<KL_INITIAL>commit/([ \t\n\r]*;|[ \t\n\r]+(try|undo|final)[^a-zA-Z_0-9]) {
	return emitNoData( KW_TrailingCommit );
}

	/* Words. */
<KL_INITIAL>{IDENT} {
	return emitToken( TK_Word, yytext, yyleng );
}

	/* Begin a shell style comment. */
<KL_INITIAL>#			{
	BEGIN(KL_SHELL_COM);
	extendGarble();
}
	/* Data in a shell style comment. */
<KL_SHELL_COM>[^\n]+		{
	extendGarble();
}
	/* Terminate a C++ style comment. */
<KL_SHELL_COM>\n		{
	BEGIN(KL_INITIAL);
	garble();
}

	/* 
	 * Start single and double literals.
	 */
<KL_INITIAL>'			{
	BEGIN(KL_SLIT);
	extendToken( 0, 0 );
}
<KL_INITIAL>\"			{
	BEGIN(KL_DLIT);
	extendToken( 0, 0 );
}

	/* Escape sequences in single and double literals. */
<KL_SLIT,KL_DLIT>\\0		extendToken( "\0", 1 );
<KL_SLIT,KL_DLIT>\\a		extendToken( "\a", 1 );
<KL_SLIT,KL_DLIT>\\b		extendToken( "\b", 1 );
<KL_SLIT,KL_DLIT>\\t		extendToken( "\t", 1 );
<KL_SLIT,KL_DLIT>\\n		extendToken( "\n", 1 );
<KL_SLIT,KL_DLIT>\\v		extendToken( "\v", 1 );
<KL_SLIT,KL_DLIT>\\f		extendToken( "\f", 1 );
<KL_SLIT,KL_DLIT>\\r		extendToken( "\r", 1 );
<KL_SLIT,KL_DLIT>\\\n		extendToken( 0, 0 );
<KL_SLIT,KL_DLIT>\\.		extendToken( yytext+1, 1 );

	/* Characters in literals. */
<KL_SLIT>[^']						extendToken( yytext, 1 );
<KL_DLIT>[^"]						extendToken( yytext, 1 );

	/* Terminate a single literal. */
<KL_SLIT>' {
	BEGIN(KL_INITIAL);
	return emitToken( TK_Literal, 0, 0 );
}
	/* Terminate a double literal */
<KL_DLIT>\" {
	BEGIN(KL_INITIAL);
	return emitToken( TK_Literal, 0, 0 );
}

	/*
	 * Whitespace. 
	 */
<KL_INITIAL>{WSCHAR}+		garble();
	
	/* Builtin code move to Builtin initial. */
<KL_INITIAL>"{" {
	/* Start an inline code block. Keep track of where it started in case
	 * it terminates prematurely. Return the open bracket. */
	BEGIN(IL_INITIAL);
	il_code_first_line = id->last_line;
	il_code_first_column = id->last_column+1;
	builtinBrace++;
	return emitNoData( *yytext );
}

	/* Any other characters. */
<KL_INITIAL>. {
	return emitNoData( *yytext );
}

	/* End of input in a literal is an error. */
<KL_SLIT,KL_DLIT><<EOF>> {
	error(id->first_line, id->first_column) << "unterminated literal" << endl;
	exit(1);
}
	
	/* End of input in a comment is an error. */
<KL_SHELL_COM><<EOF>> {
	error(id->first_line, id->first_column) << "unterminated comment" << endl;
	exit(1);
}

	/*
	 * Inline Code
	 */

	/* Begin a C style comment. */
<IL_INITIAL>"/*" {
	BEGIN(IL_C_COM);
	il_comm_lit_first_line = id->last_line;
	il_comm_lit_first_column = id->last_column+1;
	extendToken( yytext, yyleng );
}
	/* Data in a C style comment. */
<IL_C_COM>\n 	extendToken( yytext, 1 );
<IL_C_COM>. 	extendToken( yytext, 1 );

	/* Terminate a C style comment. */
<IL_C_COM>"*/" {
	BEGIN(IL_INITIAL);
	return emitToken( IL_Text, yytext, 2 );
}

	/* Begin a C++ style comment. */
<IL_INITIAL>"//" {
	BEGIN(IL_CXX_COM);
	il_comm_lit_first_line = id->last_line;
	il_comm_lit_first_column = id->last_column+1;
	extendToken( yytext, yyleng );
}
	/* Data in a C++ style comment. */
<IL_CXX_COM>[^\n]+ {
	extendToken( yytext, yyleng );
}
	/* Terminate a C++ style comment. */
<IL_CXX_COM>\n {
	BEGIN(IL_INITIAL);
	return emitToken( IL_Text, yytext, 1 );
}


	/* Start literals. */
<IL_INITIAL>' {
	BEGIN(IL_SGL_LIT);
	il_comm_lit_first_line = id->last_line;
	il_comm_lit_first_column = id->last_column+1;
	extendToken( yytext, 1 );
}
<IL_INITIAL>\" {
	BEGIN(IL_DBL_LIT);
	il_comm_lit_first_line = id->last_line;
	il_comm_lit_first_column = id->last_column+1;
	extendToken( yytext, 1 );
}
	/* Various escape sequences in literals. We don't need to get them
	 * all here. We just need to pick off the ones that could confuse us
	 * about the literal we are matching */
<IL_SGL_LIT,IL_DBL_LIT>\\'		extendToken( yytext, yyleng );
<IL_SGL_LIT,IL_DBL_LIT>\\\"		extendToken( yytext, yyleng );
<IL_SGL_LIT,IL_DBL_LIT>\\\\		extendToken( yytext, yyleng );
	/* Characters in literals. */
<IL_DBL_LIT>[^\"]				extendToken( yytext, 1 );
<IL_SGL_LIT>[^']				extendToken( yytext, 1 );

	/* Terminate a double literal */
<IL_DBL_LIT>\" {
	BEGIN(IL_INITIAL);
	return emitToken( IL_Text, yytext, 1 );
}
	/* Terminate a single literal. */
<IL_SGL_LIT>' {
	BEGIN(IL_INITIAL);
	return emitToken( IL_Text, yytext, 1 );
}

	/* Open Brace, increment count of open braces. */
<IL_INITIAL>"{" {
	builtinBrace++;
	return emitToken( IL_Text, yytext, 1 );
}

	/* Close brace, decrement count of open braces. */
<IL_INITIAL>"}" {
	builtinBrace--;
	if ( builtinBrace == 0 ) {
		/* Inline code block ends. */
		BEGIN(KL_INITIAL);
		return emitNoData( *yytext );
	}
	else {
		/* Either a semi terminated inline block or only the closing brace of
		 * some inner scope, not the block's closing brace. */
		return emitToken( IL_Text, yytext, 1 );
	}
}

	/* Words. */
<IL_INITIAL>{IDENT} {
	return emitToken( IL_Text, yytext, yyleng );
}

	/* Catch lhs ref. */
<IL_INITIAL>[\$\@][\$\@]+ {
	return emitToken( IL_Reference, yytext, yyleng );
}

<IL_INITIAL>[\$\@][0-9]+ {
	return emitToken( IL_Reference, yytext, yyleng );
}

	/* Whitespace. */
<IL_INITIAL>{WSCHAR}+ {
	return emitToken( IL_Text, yytext, yyleng );
}

	/* Any other characters. */
<IL_INITIAL>. {
	return emitToken( IL_Text, yytext, 1 );
}

<INITIAL><<EOF>> {
	/* If we are not at the bottom of the include stack, then pop the current
	 * file that we are scanning. Since we are always returning 0 to the parser
	 * it will exit and return to the parser that called it. */
	if ( inc_stack_ptr > 0 )
		popInclude();
	return 0;
}

	/* End of input in a literal is an error. */
<IL_SGL_LIT,IL_DBL_LIT><<EOF>>		{
	error(il_comm_lit_first_line, il_comm_lit_first_column) << 
			"unterminated literal" << endl;
	exit(1);
}
	
	/* End of input in a comment is an error. */
<IL_C_COM,IL_CXX_COM><<EOF>>	{
	error(il_comm_lit_first_line, il_comm_lit_first_column) <<
			"unterminated comment" << endl;
	exit(1);
}

	/* End of intput in a code block. */
<IL_INITIAL><<EOF>> {
	error(il_code_first_line, il_code_first_column) <<
			"unterminated code block" << endl;
	exit(1);
}

%%

/* Write out token data, escaping special charachters. */
#ifdef WANT_TOKEN_WRITE
void writeToken( int token, char *data )
{
	cout << "token id " << token << " at " << id->fileName << ":" <<
			yylloc->first_line << ":" << yylloc->first_column << "-" <<
			yylloc->last_line << ":" << yylloc->last_column << " ";

	if ( data != 0 ) {
		while ( *data != 0 ) {
			switch ( *data ) {
			case '\n':	cout << "\\n"; break;
			case '\t':	cout << "\\t"; break;
			default:	cout << *data; break;
			}
			data += 1;
		}
	}
	cout << endl;
}
#endif

/* Caclulate line info from yytext. Called on every pattern match. */
void updateLineInfo()
{
	/* yytext should always have at least wone char. */
	assert( yytext[0] != 0 );

	/* Scan through yytext up to the last character. */
	char *p = yytext;
	for ( ; p[1] != 0; p++ ) {
		if ( p[0] == '\n' ) {
			id->last_line += 1;
			id->last_column = 0;
		}
		else {
			id->last_column += 1;
		}
	}

	/* Always consider the last character as not a newline. Newlines at the
	 * end of a token are as any old character at the end of the line. */
	id->last_column += 1;

	/* The caller may be about to emit a token, be prepared to pass the line
	 * info to the parser. */
	yylloc->first_line = id->first_line;
	yylloc->first_column = id->first_column;
	yylloc->last_line = id->last_line;
	yylloc->last_column = id->last_column;

	/* If the last character was indeed a newline, then wrap ahead now. */
	if ( p[0] == '\n' ) {
		id->last_line += 1;
		id->last_column = 0;
	}
}

/* Extend the garbling of a matched pattern that will not be part of a token.
 * Provided so that line info of garbled tokens can be preserved across
 * multiple patterns. */
void extendGarble() 
{
	/* Update line information from yytext. */
	updateLineInfo();
}

/* Eat up a matched pattern that will not be part of a token. */
void garble() 
{
	/* Update line information from yytext. */
	updateLineInfo();

	/* The next token starts ahead of the last token. */
	id->first_line = id->last_line;
	id->first_column = id->last_column + 1;
}

/* Append data to the end of the token. More token data expected. */
void extendToken( char *data, int len )
{
	if ( data != 0 && len > 0 )
		tokbuf.append( data, len );

	/* Update line information from yytext. */
	updateLineInfo();
}

/* Possibly process include data. */
void processInclude( int token )
{
	static char *incFileName = 0;

	if ( handlingInclude ) {
		if ( token == KW_Include )
			incFileName = 0;
		else if ( token == TK_Literal )
			incFileName = yylval->data;
		else if ( token == ';' ) {
			/* Terminate the include statement. Start reading from included file. */
			handlingInclude = false;

			if ( id->active && includeDepth < INCLUDE_STACK_SIZE ) {
				/* If there is no section name or input file, default to the curren values. */
				if ( incFileName == 0 )
					incFileName = id->fileName;

				/* Make the new buffer and switch to it. */
				FILE *incFile = fopen( incFileName, "rt" );
				if ( incFile != 0 ) {
					buff_stack[inc_stack_ptr++] = YY_CURRENT_BUFFER;
					yy_switch_to_buffer( yy_create_buffer( incFile, YY_BUF_SIZE ) );
					BEGIN(INITIAL);
				}
				else {
					error(*yylloc) << "could not locate include file \"" << incFileName 
							<< "\"" << endl;
				}
			}
		}
	}
}

void popInclude()
{
	/* Free the current buffer and move to the previous. */
	yy_delete_buffer( YY_CURRENT_BUFFER );
	yy_switch_to_buffer( buff_stack[--inc_stack_ptr] );

	/* Includes get called only from KL_INITIAL. */
	BEGIN(KL_INITIAL);
}


/* Append data to the end of a token and emitToken it to the parser. */
int emitToken( int token, char *data, int len )
{
	/* Append the data and null terminate. */
	if ( data != 0 && len > 0 )
		tokbuf.append( data, len );
	tokbuf.append( 0 );

	/* Duplicate the buffer. */
	yylval->data = new char[tokbuf.length];
	strcpy( yylval->data, tokbuf.data );

	/* Update line information from yytext. */
	updateLineInfo();

	/* Write token info. */
#ifdef WANT_TOKEN_WRITE
	writeToken( token, tokbuf.data );
#endif

	/* Clear out the buffer. */
	tokbuf.clear();

	/* The next token starts ahead of the last token. */
	id->first_line = id->last_line;
	id->first_column = id->last_column + 1;

	/* Possibly process the include statement; */
	processInclude( token );

	return token;
}

/* Emit a token with no data to the parser. */
int emitNoData( int token ) 
{
	/* Return null to the parser. */
	yylval->data = 0;

	/* Update line information from yytext. */
	updateLineInfo();

	/* Write token info. */
#ifdef WANT_TOKEN_WRITE
	writeToken( token, 0 );
#endif

	/* Clear out the buffer. */
	tokbuf.clear();

	/* The next token starts ahead of the last token. */
	id->first_line = id->last_line;
	id->first_column = id->last_column + 1;

	/* Possibly process the include statement; */
	processInclude( token );

	return token;
}

/* Pass tokens in outter code through to the output. */
void passThrough( char *data )
{
	/* If no errors, we are emitting code and we are at the bottom of the
	 * include stack (the source file listed on the command line) then write
	 * out the data. */
	if ( !generateGraphviz && gblErrorCount == 0 && inc_stack_ptr == 0 )
		*outStream << data;
}

/* Init a buffer. */
Buffer::Buffer() 
:
	data(0), 
	length(0),
	allocated(0)
{
}

/* Empty out a buffer on destruction. */
Buffer::~Buffer()
{
	empty();
}

/* Free the space allocated for the buffer. */
void Buffer::empty()
{
	if ( data != 0 ) {
		free( data );

		data = 0;
		length = 0;
		allocated = 0;
	}
}

/* Grow the buffer when to len allocation. */
void Buffer::upAllocate( int len )
{
	if ( data == 0 )
		data = (char*) malloc( len );
	else
		data = (char*) realloc( data, len );
	allocated = len;
}

int yywrap()
{
	/* Once processessing of the input is done, signal no more. */
	return 1;
}

/* Here simply to suppress the unused yyunpt warning. */
void thisFuncIsNeverCalled()
{
	yyunput(0, 0);
}

/* Put the scannner back into the outside code start state. */
void beginOutsideCode()
{
	BEGIN(INITIAL);
}

/* Wrapper for the lexer which stores the locations of the value and location
 * variables of the parser into globals. The parser is reentrant, however the scanner
 * does not need to be, so globals work fine. This saves us passing them around
 * all the helper functions. */
int yylex( YYSTYPE *yylval, YYLTYPE *yylloc )
{
	::yylval = yylval;
	::yylloc = yylloc;
	return kelbt_lex();
}

